<?php

declare(strict_types=1);

namespace Reloadify\Service;

use Doctrine\DBAL\Connection;
use Doctrine\DBAL\Exception;
use Reloadify\Entity\AbandonedCart\AbandonedCartDefinition;
use Reloadify\Entity\AbandonedCart\AbandonedCartEntity;
use Shopware\Core\Checkout\Cart\Cart;
use Shopware\Core\Checkout\Cart\Error\ErrorCollection;
use Shopware\Core\Checkout\Cart\RedisCartPersister;
use Shopware\Core\Defaults;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\Doctrine\RetryableQuery;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\Struct\ArrayStruct;
use Shopware\Core\Framework\Uuid\Uuid;

class AbandonedCartService
{
    private Connection         $connection;
    protected EntityRepository $abandonedCartRepository;

    /**
    * param cannot be natively typed, as symfony might change the type in the future
    *
    * @param \Redis|\RedisArray|\RedisCluster|\Predis\ClientInterface|\Relay\Relay $redis
    */
    private $redis;

    public function __construct(
        Connection $connection,
        EntityRepository $abandonedCartRepository,
        $redis
    )
    {
        $this->connection = $connection;
        $this->abandonedCartRepository = $abandonedCartRepository;
        $this->redis = $redis;
    }

    /**
     * @param string  $id
     * @param Context $context
     * @return AbandonedCartEntity|null
     */
    public function getAbandonedCartWithId(string $id, Context $context): ?AbandonedCartEntity
    {
        $criteria = new Criteria([$id]);
        /** @var AbandonedCartEntity */
        return $this->abandonedCartRepository->search($criteria, $context)->first();
    }

    /**
     * @param string  $token
     * @param Context $context
     * @return AbandonedCartEntity|null
     */
    public function getAbandonedCartWithToken(string $token, Context $context): ?AbandonedCartEntity
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::TOKEN_FIELD, $token));

        /** @var AbandonedCartEntity */
        return $this->abandonedCartRepository->search($criteria, $context)->first();
    }

    public function setAbandonedCartToken(string $id, string $token, Context $context): void
    {
        $this->abandonedCartRepository->update(
            [
                [
                    'id'                                 => $id,
                    AbandonedCartDefinition::TOKEN_FIELD => $token,
                ],
            ],
            $context
        );
    }

    /**
     * @param Context $context
     * @return void
     */
    public function deleteOldAbandonedCarts(Context $context): void
    {
        $criteria = new Criteria();
        $date = new \DateTime();
        $date->sub(new \DateInterval('P30D'));
        $criteria->addFilter(
            new RangeFilter(
                'createdAt',
                [RangeFilter::LTE => $date->format(Defaults::STORAGE_DATE_TIME_FORMAT)]
            )
        );

        $carts = $this->abandonedCartRepository->searchIds($criteria, $context);
        $ids = \array_map(
            static function ($id) {
                return ['id' => $id];
            },
            $carts->getIds()
        );
        $this->abandonedCartRepository->delete(
            $ids,
            $context
        );
    }

    /**
     * @param string  $token
     * @param Context $context
     * @return void
     */
    public function deleteAbandonedCart(string $token, Context $context): void
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::TOKEN_FIELD, $token));

        $abandonedCartId = $this->abandonedCartRepository->searchIds($criteria, $context)->firstId();

        if (empty($abandonedCartId)) {
            return;
        }

        $this->abandonedCartRepository->delete(
            [
                [
                    'id' => $abandonedCartId,
                ],
            ],
            $context
        );
    }

    public function deleteAbandonedCartByCustomerId(string $customerId, Context $context): void
    {
        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::CUSTOMER_ID_FIELD, $customerId));

        $abandonedCartId = $this->abandonedCartRepository->searchIds($criteria, $context)->firstId();

        if (empty($abandonedCartId)) {
            return;
        }

        $this->abandonedCartRepository->delete(
            [
                [
                    'id' => $abandonedCartId,
                ],
            ],
            $context
        );
    }

    /**
     * @throws Exception
     */
    public function saveAbandonedCart(Cart $cart, string $token, string $customerId, string $salesChannelId, string $languageId, string $currencyId, float $totalPrice, float $netPrice, Context $context): void
    {
        if(!empty($this->redis)){
            $value = $this->redis->get(RedisCartPersister::PREFIX . $token);

            if ($value === false || !\is_string($value)) {
                $this->deleteAbandonedCart($token, $context);
                return;
            }

            try {
                $value = \unserialize($value);
            } catch (\Exception) {
                $this->deleteAbandonedCart($token, $context);
                return;
            }

            $cartData = [];
            $cartData['created_at'] = $cart->getExtension('createdAt')->getVars()[0];
            $cartData['updated_at'] = $cart->getExtension('updatedAt')->getVars()[0];
        }else {
            $sql = <<<SQL
SELECT `created_at` FROM `cart` WHERE `token` = :token;
SQL;

            $result = $this->connection->executeQuery($sql, ['token' => $token]);
            $cartData = $result->fetchAssociative();

            if(!isset($cartData["created_at"])){
                $cartData['created_at'] = $cart->getExtension('createdAt')->getVars()[0];

            }

            $cartData['updated_at'] = $cart->getExtension('updatedAt')->getVars()[0];
        }

        if($cartData === false) {
            // Could not fetch data from the actual cart using this token, so remove the abandoned cart as well.
            $this->deleteAbandonedCart($token, $context);
            return;
        }

        $criteria = new Criteria();
        $criteria->addFilter(new EqualsFilter(AbandonedCartDefinition::TOKEN_FIELD, $token));
        $payload = $cart->getLineItems()->getElements();

        $abandonedCartId = $this->abandonedCartRepository->searchIds($criteria, $context)->firstId();
        if (empty($abandonedCartId)) {
            $abandonedCartId = Uuid::randomBytes();
        }
        else {
            $abandonedCartId = Uuid::fromHexToBytes($abandonedCartId);
        }
        $this->connection->beginTransaction();

        $sql = <<<'SQL'
        INSERT INTO reloadify_abandoned_cart (id, token, cart, payload, customer_id, sales_channel_id, language_id, currency_id, amount_total, amount_net, created_at)
        VALUES (:id, :token, :cart, :payload, :customer_id, :sales_channel_id, :language_id, :currency_id, :amount_total, :amount_net, :created_at)
        ON DUPLICATE KEY UPDATE
            id = :id,
            token = :token,
            cart = :cart,
            payload = :payload,
            customer_id = :customer_id,
            sales_channel_id = :sales_channel_id,
            language_id = :language_id,
            currency_id = :currency_id,
            amount_total = :amount_total,
            amount_net = :amount_net,
            updated_at = :updated_at;
        SQL;

        $serializedCart = $this->serializeCart($cart);

        $data = [
            'id'          => $abandonedCartId,
            'token'       => $token,
            'cart'        => $serializedCart,
            'payload'     => json_encode($payload),
            'customer_id' => Uuid::fromHexToBytes($customerId),
            'sales_channel_id' => Uuid::fromHexToBytes($salesChannelId),
            'language_id' => Uuid::fromHexToBytes($languageId),
            'currency_id' => Uuid::fromHexToBytes($currencyId),
            'amount_total' => $totalPrice,
            'amount_net' => $netPrice,
            ...$cartData
        ];

        try {
            $query = new RetryableQuery($this->connection, $this->connection->prepare($sql));

            $query->execute($data);
        }
        catch (\Throwable $e) {
            //Do nothing
        }

        $this->connection->commit();
    }

    public function addCreatedAt(Cart $cart): void
    {
        $cart->addExtension('createdAt',  new ArrayStruct([(new \DateTime())->format(Defaults::STORAGE_DATE_TIME_FORMAT)]));
        $cart->addExtension('updatedAt',  new ArrayStruct([(new \DateTime())->format(Defaults::STORAGE_DATE_TIME_FORMAT)]));
    }

    private function serializeCart(Cart $cart): string
    {
        $errors = $cart->getErrors();
        $data = $cart->getData();

        $cart->setErrors(new ErrorCollection());
        $cart->setData(null);

        $serializedCart = serialize($cart);

        $cart->setErrors($errors);
        $cart->setData($data);

        return $serializedCart;
    }

}
